#pragma once

#include <filesystem>
#include <fstream>
#include <omp.h>
#include <map>
#include <iostream>

#include "csvol_constants.h"
#include "VolumeCompressionBase.hpp"
#include "rans.hpp"
#include "util.hpp"
#include "space_filling_curves.hpp"

// How many CPU Threads to use to parallelize encoding/decoding etc.
// Usually, the number of available CPU cores is a good fit.
#define NUM_THREADS 8

namespace vvv {

class CompressedSegmentationVolume : public VolumeCompressionBase {

public:
    explicit CompressedSegmentationVolume(uint32_t brick_size=0) : VolumeCompressionBase(), m_brick_size(brick_size), m_encoding(),
                                                                   m_brick_starts(), m_volume_dim(-1), m_use_rANS(false), m_use_detail_freq(false) {}

    // COMPRESSION -----------------------------------------------------------------------------------------------------

    /**
     * Sets the options for the compression step. If using rANS, 64 bit frequency tables must be given as uint32_t[16] arrays.
     */
    void setCompressionOptions(bool use_rANS, bool use_detail_freq, const uint32_t* operation_frequencies = nullptr, const uint32_t* detail_operation_frequencies = nullptr);

    /**
     * Sets the options for the compression step. If using rANS, 64 bit frequency tables must be given as size_t[16] arrays.
     */
    void setCompressionOptions64(bool use_rANS, bool use_detail_freq, const size_t* operation_frequencies = nullptr, const size_t* detail_operation_frequencies = nullptr);

    /**
     * Performs a pseudo compression pass to obtain operation frequency tables for later rANS encoding.
     * @param detail_freq if true, a separate table is computed for the finest LoD.
     * @param freq_out first 16 entries contain the base frequencies. If detail_freq, the latter 16 elements contain the table for the finest LoD.
     * @param subsampling_factor if > 1, only every other brick is handled. The higher the factor, the fewer bricks are processed.
     */
    void compressForFrequencyTable(const std::vector<uint32_t>& volume, glm::uvec3 volume_dim, size_t freq_out[32], uint32_t subsampling_factor, bool detail_freq, bool verbose=false);

    /**
     * Converts the 64 bit operation frequency table into a 32 bit table with identical relative operation frequencies,
     * normalized for rANS encoding. Any operation frequency that is zero is set to a minimal value of 1. This is due
     * to the potential subsampling of bricks when computing the frequencies which could miss counting certain operations
     * in small data sets. Later rANS encoding behavior would be undefined when encoding operations with zero frequencies.
     */
    static std::vector<uint32_t> normalizeOperationFrequencies(const size_t* freq);

    /**
     * Normalizes this frequency table for rANS encoding. Any operation frequency that is zero is set to a minimal value
     * of 1. This is due to the potential subsampling of bricks when computing the frequencies which could miss counting
     * certain operations in small data sets. Later rANS encoding behavior would be undefined when encoding operations
     * with zero frequencies.
     */
    static std::vector<uint32_t> normalizeOperationFrequencies(const uint32_t* freq);

    /**
     * Compresses the given volume for the previously set compression options and stores the result internally.
     * Afterwards, single bricks or the full volume can be decoded to any LoD.
     * @param volume original label volume with 32 bit labels and voxels stored in XYZ linearized order.
     * @param volume_dim dimensions in voxels of the original volume.
     */
    void compress(const std::vector<uint32_t>& volume, glm::uvec3 volume_dim, bool verbose=false) override;

    // DECOMPRESSION ---------------------------------------------------------------------------------------------------

    /**
     * Decompresses a single brick to the given output buffer. Note that the voxels will be in morton order in out!
     * Additional remapping is required if another linearization should be present.
     * @param out buffer with at least brick_size^3 elements
     * @param brick the brick to decompress
     * @param target_lod the target LoD to compress the brick to
     */
    void decompressBrickTo(uint32_t* out, glm::uvec3 brick, int target_lod) const;

    /**
     * Decompresses the full volume up to a certain LoD into the vector out.
     */
    void decompressLoD(int lod, std::vector<uint32_t>& out) const;

    /**
     * Decompresses the full original volume an returns it as a new uint32_t vector.
     */
    std::shared_ptr<std::vector<uint32_t>> decompress() override {
        std::shared_ptr<std::vector<uint32_t>> out = std::make_shared<std::vector<uint32_t>>();
        out->resize(static_cast<size_t>(m_volume_dim.x) * m_volume_dim.y * m_volume_dim.z);
        MiniTimer t;
        decompressLoD(0, *out);
        double time = t.elapsed();
        std::cout << " full decode in " << t.elapsed() << "s (" << (static_cast<double>(out->size() * sizeof(uint32_t))/time/1000/1000/1000) << " GB/s)";
        return out;
    }

    // TESTS -----------------------------------------------------------------------------------------------------------

    /**
     * Checks if all LoDs except the finest one are decompressed correctly.
     * Any node should have the same label as given in the reference multigrids.
     */
    bool testLoD(const std::vector<uint32_t>& volume, glm::uvec3 volume_dim) const;

    /**
     * Tests if the original volume can be reconstructed without errors from the encoding and if all available LoDs
     * can be reconstructed as defined by the reference multigrids per brick.
     * @param compress_first if true, the volume is compressed before testing
     */
    bool test(const std::vector<uint32_t>& volume, const glm::uvec3 volume_dim, bool compress_first=false) override {
        if(!VolumeCompressionBase::test(volume, volume_dim, compress_first)) {
            std::cout << "skipping LODs...";
            std::cout << "-------------------------------------------------------------";
            return false;
        }
        return testLoD(volume, volume_dim);
    }

    /**
     * Linearization of all 3D brick indices. All compressed bricks are stored along this order.
     * Also used to access the start indices in the brick_starts array.
     */
    static inline uint32_t brick_to_1D(glm::uvec3 brick_pos, glm::uvec3 brick_count) {
        return sfc::Cartesian::p2i(brick_pos, brick_count);
    }
    bool isUsingRANS() const { return m_use_rANS; }
    bool isUsingDetailFreq() const { return m_use_detail_freq; }
    const std::vector<uint32_t>* getEncoding() { return &m_encoding; }
    const std::vector<uint32_t>* getBrickStarts() { return &m_brick_starts; }
    glm::uvec3 getVolumeDim() const { return m_volume_dim; }
    glm::uint32_t getBrickSize() const { return m_brick_size; }
    uint32_t getLodCountPerBrick() const { return static_cast<uint32_t>(log2(m_brick_size)) + 1; }
    glm::uvec3 getBrickCount() const { return (m_volume_dim - glm::uvec3(1)) / m_brick_size + 1u; }


    // FILE IMPORT/EXPORT ----------------------------------------------------------------------------------------------
    static std::string getCSVFileName(const std::string& filepath, uint32_t brick_size, bool use_rANS,
                                      bool use_detail_freq, const std::string filetype= ".cmp") {
        return filepath.substr(0, filepath.rfind('.')) + "_difftree_bs" + std::to_string(brick_size) +
                    (use_rANS ? "_rANS" : "") + (use_detail_freq ? "df" : "") + filetype;
    }
    std::string getCSVFileName(const std::string& filepath, const std::string filetype= ".cmp") {
        return getCSVFileName(filepath, m_brick_size, m_use_rANS, m_use_detail_freq, filetype);
    }
    void exportToFile(const std::string& path, bool verbose = true);
    bool importFromFile(const std::string& path, bool verbose = true);

    // UTILITY ---------------------------------------------------------------------------------------------------------
    void clear() {
        m_volume_dim = glm::uvec3(0u);
        m_brick_size = 0u;
        m_encoding.clear();
        m_brick_starts.clear();
    }

    float getCompressedSizeInGB() {
        return static_cast<float>(m_encoding.size() + m_brick_starts.size()) * sizeof(uint32_t) / 1000.f / 1000.f / 1000.f;
    }

    /**
     * Returns a compression ratio measured as [compressed size] / [original uint32 data set size] in percent.
     */
    float getCompressionRatio() override {
        return static_cast<float>((m_encoding.size() + m_brick_starts.size()) * sizeof(uint32_t)) /
               static_cast<float>(m_volume_dim.x * m_volume_dim.y * m_volume_dim.z * sizeof(uint32_t)) * 100.f;
    }

    std::string decodingInfoString() {
        const float brick_starts_memory = static_cast<float>(m_brick_starts.size() * sizeof(uint32_t)) / 1000.f / 1000.f;
        const float encoding_memory = static_cast<float>(m_encoding.size() * sizeof(uint32_t)) / 1000.f / 1000.f;
        const float volume_memory = static_cast<float>(static_cast<size_t>(m_volume_dim[0] * m_volume_dim[1] * m_volume_dim[2]) * sizeof(uint32_t)) / 1000.f / 1000.f;
        std::stringstream ss;
        ss << "start buffer (" << brick_starts_memory << "MB) + encoding buffer (" << encoding_memory << "MB) = "
           << (brick_starts_memory + encoding_memory) << "MB / " << volume_memory << "MB original size ("
           << (static_cast<float>(brick_starts_memory + encoding_memory) / volume_memory*100.f) << "%)";
        return ss.str();
    }


private:
    // list of neighbor vectors per index (8 indices in total) where each index has 3 ivec3 vectors of type [-1, 1]^3
    static constexpr const glm::ivec3 neighbor[][3] = {{glm::ivec3({-1, 0, 0}), glm::ivec3({0, -1, 0}), glm::ivec3({0, 0, -1})},
                                                       {glm::ivec3({1, 0, 0}), glm::ivec3({0, -1, 0}), glm::ivec3({0, 0, -1})},
                                                       {glm::ivec3({-1, 0, 0}), glm::ivec3({0, 1, 0}), glm::ivec3({0, 0, -1})},
                                                       {glm::ivec3({1, 0, 0}), glm::ivec3({0, 1, 0}), glm::ivec3({0, 0, -1})},
                                                       {glm::ivec3({-1, 0, 0}), glm::ivec3({0, -1, 0}), glm::ivec3({0, 0, 1})},
                                                       {glm::ivec3({1, 0, 0}), glm::ivec3({0, -1, 0}), glm::ivec3({0, 0, 1})},
                                                       {glm::ivec3({-1, 0, 0}), glm::ivec3({0, 1, 0}), glm::ivec3({0, 0, 1})},
                                                       {glm::ivec3({1, 0, 0}), glm::ivec3({0, 1, 0}), glm::ivec3({0, 0, 1})}};

    /**
     * Returns the current value in the brick at the neighbor_i neighbor position of brick_pos at the decoding stage at the given lod_width.
     * If the neighbor is not yet set in this level, the parent element of this neighbor is returned.
     * If the neighbor would lie outside the brick, INVALID is returned.
     */
    static uint32_t valueOfNeighbor(const uint32_t* brick, const glm::uvec3& brick_pos, uint32_t local_lod_i, uint32_t lod_width, uint32_t brick_size, int neighbor_i);
    static uint32_t valueOfNeighbor(const MultiGridNode* grid, const MultiGridNode* parent_grid, const glm::uvec3& brick_pos, uint32_t local_lod_i, uint32_t lod_width, int neighbor_i);

    /**
     * Enumerate the voxel positions within a brick.
     * Because of how we encode the LoDs, this enumeration is required to always be in an "octree manner":
     * Iterating over it with a step size of 2*2*2=8 should land on all start points of 2x2x2 bricks in the Octree and so on.
     * Morton and Hilbert curves for example satisfy this criterion.
     */
    static inline glm::uvec3 enumBrickPos(uint32_t i, uint32_t brick_size) { return sfc::Morton3D::i2p(i); }
    static inline glm::uint32_t indexOfBrickPos(const glm::uvec3& p) { return sfc::Morton3D::p2i(p); }

    /**
     * Computes operation frequencies and detail operation frequencies (the latter offsetted by 16) for the brick into the given brick_freq[32] array.
     */
    void compressBrickForFrequencyTable(const std::vector<uint32_t>& volume, size_t* brick_freq, glm::uvec3 start, glm::uvec3 volume_dim, bool detail_freq) const;


    /**
     * Encodes a single brick from given start with size brick_size in the volume to the out vector. out should have enough space reserved for adding all elements.
     * @return number of uint32_t elements written to out
     */
    uint32_t compressBrick(const std::vector<uint32_t>& volume, std::vector<uint32_t>& out, glm::uvec3 start, glm::uvec3 volume_dim);

    /**
     * Uses the given rANS encoder to compress the stream of nibbles between the given start and end point.
     */
    static uint32_t packRANS(std::vector<uint32_t>& v, uint32_t start4bit, uint32_t end4bit, RANS& rANS);

    struct ReadState {
        uint32_t idxE = 0u;         // Used either as 4 bit element index or byte index for rANS.
        uint32_t rans_state = 0u;   // This is the whole rANS decoding state.
        bool in_detail_lod = false; // Whether we're currently reading from the finest LoD.

    };
    /**
     * Reads the next element from the encoding, possibly using the rANS decoder, and updates the state.
     */
    uint32_t readNextLodOperation(size_t start32bit, CompressedSegmentationVolume::ReadState &state) const;

    /**
     * Decompresses a single brick in the volume. A volume has to be compressed beforehand.
     * The brick_idx is used to read the begin and endpoint of the encoding from the brick_starts buffer.
     * Output is an uint32_t array of the decoded brick. It always has to have brick_size^3 elements, but output
     * values for elements outside the volume dimension are undefined. Output elements are in morton order.
     *
     * @param valid_brick_size usually brick_size^3 except for bricks at the border of the volume. Has to specify the dimension of the potentially clamped brick in the original volume.
     * @param inv_lod the inverse LoD until which to decompress, or rather, the decompression iterations. 0 is the coarsest and log2(brick_size) is the original / finest level.
     */
    void decompressBrick(uint32_t brick_idx, uint32_t* output_brick, glm::uvec3 valid_brick_size, int inv_lod) const;

    uint32_t m_brick_size;                  // size of one axis, must be power of two. Each brick spans brick_size^3 voxels.
    glm::uvec3 m_volume_dim;                // size of the input volume to compress.
    std::vector<uint32_t> m_encoding;       // contains all compressed entries for all LoDs of all bricks.
    std::vector<uint32_t> m_brick_starts;   // contains start index of each brick in m_encoding.

    RANS m_rans;                            // rANS encoder with a given frequency table.
    RANS m_detail_rans;                     // rANS encoder with another frequency table for the finest LoD.

    bool m_use_rANS;                        // if rANS encoding should be used as a secondary compression.
    bool m_use_detail_freq;                 // if the rANS encoding should use a second frequency table for the finest LoD.
};

} // namespace vvv
