#pragma once

#include <cstdint>
#include <vector>
#include <memory>

#include <glm/glm.hpp>
#include <cstdlib>
#include <iostream>
#include <fstream>
#include <sstream>

class Volume {

public:
    Volume() : dim_x(0u), dim_y(0u), dim_z(0u), m_data() {}

    uint32_t dim_x = 0u, dim_y = 0u, dim_z = 0u;
    const std::vector<uint32_t>& data() const { return m_data; }

    /**
     * Replaces this volume with a Voronoi-like volume with the specified width, height, and depth and approx.
     * points_per_million_voxels labels per million voxels. Internally, a poisson point set is created before computing
     * nearest neighbors within this point set for all voxels in the volume. Generating large volumes with many labels
     * can lead to high compute times.
     */
    void createSyntheticSegmentationVolume(int w, int h, int d, size_t points_per_million_voxels = 100, size_t candidate_count = 16) {
        const size_t point_count = static_cast<size_t>(static_cast<float>(w * h * d) / 1000000.f
                                                        * static_cast<float>(points_per_million_voxels));
        std::vector<glm::ivec3> candidates(candidate_count);
        std::vector<uint32_t> dist2(candidate_count);
        uint32_t max_candidate;

        // Generated poisson-like point cloud
        std::vector<glm::ivec3> points(point_count);
        for (size_t i = 0; i < point_count; i++) {
            // Create candidate_count candidates and keep the candidate point with the maximum distance to all existing
            // points in the data set.
            #pragma omp parallel for default(none) shared(candidates, dist2, candidate_count, w, h, d, points, i)
            for (size_t n = 0; n < candidate_count; n++) {
                candidates[n].x = std::rand() % static_cast<int>(w);
                candidates[n].y = std::rand() % static_cast<int>(h);
                candidates[n].z = std::rand() % static_cast<int>(d);
                dist2[n] = -1u;
                glm::ivec3 dist;
                uint32_t d2;
                // Careful with large volumes / point sets: the following step can take ages.
                // A better implementation would use an acceleration structure to speed up the nearest neighbor search.
                for (size_t p = 0; p < i; p++) {
                    dist = points[p] - candidates[n];
                    d2 = dist.x * dist.x + dist.y * dist.y + dist.z * dist.z;
                    if (d2 < dist2[n]) {
                        dist2[n] = d2;
                    }
                }
            }
            max_candidate = 0;
            for (uint32_t n = 1; n < candidate_count; n++) {
                if (dist2[n] > dist2[max_candidate]) {
                    max_candidate = n;
                }
            }
            points[i] = candidates[max_candidate];

            if (i % (point_count / 100) == 0) {
                std::cout << "Poisson point cloud creation: " << (100 * i / point_count) << "%"
                          << " of " << point_count << " points." <<"\r" << std::flush;
            }
        }
        std::cout << "Filling voxels with nearest neighbors..                  " << "\r" << std::flush;

        // now we have our nice point set. Next step: create a voronoi volume from it
        dim_x = w;
        dim_y = h;
        dim_z = d;
        m_data.resize(w * h * d);
        #pragma omp parallel for collapse(3) default(none) shared(points, w, h, d)
        for (int z = 0; z < d; z++) {
            for (int y = 0; y < h; y++) {
                for (int x = 0; x < w; x++) {
                    glm::ivec3 pos(x, y, z);
                    glm::ivec3 dist = pos - points[0];
                    uint32_t d2 = dist.x * dist.x + dist.y * dist.y + dist.z * dist.z;
                    uint32_t min_point = 0;
                    uint32_t min_d2 = d2;
                    // And again, the following step can take a long time for large volumes:
                    for (uint32_t i = 1; i < points.size(); i++) {
                        dist = pos - points[i];
                        d2 = dist.x * dist.x + dist.y * dist.y + dist.z * dist.z;
                        if (d2 < min_d2) {
                            min_point = i;
                            min_d2 = d2;
                        }
                    }
                    m_data[z * (dim_x * dim_y) + y * dim_z + x] = min_point;
                }
            }
        }
    }


    // Volumes can be loaded and stored in very simple NRRD-like format as [two lines header, raw payload]:
    // "[volume width] [volume height] [volume depth]"
    // "[volume data type]"
    // "[raw-array of data]"
    void exportToFile(std::string out_path) {
        std::ofstream fout(out_path, std::ios::out);
        if(!fout.is_open()) {
            std::cout << "Could not open volume export file " << out_path << "! Aborting." << std::endl;
        }
        fout << dim_x << " " << dim_y << " " << dim_z << std::endl;
        fout << "uint32" << std::endl;
        fout.write(reinterpret_cast<char *>(m_data.data()), m_data.size() * sizeof(uint32_t));
        fout.close();
    }

    bool importFromFile(std::string in_path) {
        std::ifstream fin(in_path, std::ios::out);
        if(!fin.is_open()) {
            std::cout << "Could not open volume import file " << in_path << "! Aborting." << std::endl;
            return false;
        }

        std::string line;
        // first line contains space seperated width height depth
        if (!std::getline(fin, line)) {
            fin.close();
            std::cout << "Unexpected end of file in " << in_path << "! Aborting." << std::endl;
            return false;
        }
        std::istringstream sizes(line);
        sizes >> dim_x >> dim_y >> dim_z;
        // second line contains data type
        if (!std::getline(fin, line)) {
            fin.close();
            std::cout << "Unexpected end of file in " << in_path << "! Aborting." << std::endl;
            return false;
        }
        if (line != "uint32") {
            fin.close();
            std::cout << "Data format " << line << " is not supported. Expected uint32! Aborting." << std::endl;
            return false;
        }

        const size_t voxel_count = dim_x * dim_y * dim_z;
        m_data.resize(voxel_count);
        fin.read(reinterpret_cast<char *>(m_data.data()), static_cast<long>(voxel_count * sizeof(uint32_t)));
        if (!fin) {
            std::cout << "Could only read " << std::to_string(fin.gcount()) << " of expected "
                      << std::to_string(voxel_count * sizeof(uint32_t)) << "bytes!";
            fin.close();
            return false;
        }
        fin.close();
        return true;
    }

private:
    std::vector<uint32_t> m_data = {};

};