#pragma once

// adapted form the public domain ryg_rans rANS code by Fabian Giesen.
// see README.txt in include/rANS for details.

#include <cassert>
#include <cstring>
#include <sstream>
#include <string>
#include <vector>

#include <glm/glm.hpp>

#include "rANS/rans_nibble.h"

namespace vvv {

#define RANS_ALPHABET_SIZE 16

// ---- Stats
struct SymbolStats {
    uint32_t freqs[RANS_ALPHABET_SIZE];
    uint32_t cum_freqs[RANS_ALPHABET_SIZE + 1];

    void count_freqs(uint8_t const *in, size_t nbytes);
    void calc_cum_freqs();
    void normalize_freqs(uint32_t target_total);
};


class RANS {

private:
    static const uint32_t prob_bits = 14;
    static const uint32_t prob_scale = 1 << prob_bits;

    // Big01 frequencies
    // WITH RLE
    //static constexpr uint32_t _fixed_freq[RANS_ALPHABET_SIZE] = {48202,17346,13600,8479,5961,7251,3676,35787,164102,82383,37653,17155,4946,2766,2898,187};

    RansEncSymbol esyms[RANS_ALPHABET_SIZE];
    RansDecSymbol dsyms[RANS_ALPHABET_SIZE];
    uint8_t cum2sym[prob_scale];
    SymbolStats stats;
    bool has_frequency_tables = false;

public:
    RANS(const uint32_t *frequency_array = nullptr) {
        if(frequency_array)
            recomputeFrequencyTables(frequency_array);
    }

    void recomputeStaticFrequencyTables(bool rle, bool palette_delta, bool mtf) {
        int index = 0 + (rle ? 1 : 0) + (palette_delta ? 2 : 0) + (mtf ? 4 : 0);
        constexpr uint32_t freq[8][RANS_ALPHABET_SIZE] = {{34348398,12353221,6980536,3731146,0,2588286,583112,0,292140797,87522960,31339420,10214634,0,1706446,914476,0}, // NO RLE, NO DELTA, NO MTF x
                                                          {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, // RLE,    NO DELTA, NO MTF
                                                          {34622857,12500575,7070347,3797499,367003,2347588,596492,0,292169364,87550125,31364824,10238574,633648,1286117,767857,20234}, // NO RLE, DELTA,    NO MTF x
                                                          {48202,17346,13600,8479,5961,7251,3676,35787,164102,82383,37653,17155,4946,2766,2898,187}, // RLE,    DELTA,    NO MTF x
                                                          {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, // NO RLE, NO DELTA, MTF
                                                          {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, // RLE,    NO DELTA, MTF
                                                          {209313281,152575121,68505625,27940834,11289632,6696073,3922867,2271703,1201771,668477,398199,254278,168337,91472,30166,5268}, // NO RLE, DELTA,    MTF x
                                                          {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}};// RLE,    DELTA,    MTF
        assert(freq[index][0] > 0 && " this frequency table does not yet exist! please recompute the compression without rANS and copy the frequency table to rans.hpp");
        recomputeFrequencyTables(freq[index]);
        // for hardcoded Fiber Polymer data set:
//        constexpr uint32_t freq_poly[RANS_ALPHABET_SIZE] = {24933215,5479512,2757734,1022971,805234,328002,426147,0,184334129,46344079,12791205,1448358,391163,85886,184283,31258};
//        recomputeFrequencyTables(freq_poly);
    }
    void recomputeFrequencyTables(const uint32_t * frequency_array = nullptr);
    void recomputeFrequencyTables(std::vector<uint8_t> &in_bytes);

    void copyCurrentFrequencyTableTo(uint32_t * frequency_array) {
        assert(has_frequency_tables && "can't copy frequency table because it doesn't exist");
        assert(frequency_array && "invalid pointer to write buffer for frequency array");
        for(int i = 0; i < 16; i++)
            frequency_array[i] = stats.freqs[i];
    }

    /**
     * Replaces all 4 bit elements between start4bit (including) and end4bit (excluding)  in_packed with a rANS encoded bytestream.
     * The first 4 bit element start4bit must be the first position in a 32bit memory location.
     * @return the new end4bit endpoint measured in number of 4 bit elements
     */
     uint32_t packRANS(std::vector<uint32_t> &in_packed, uint32_t start4bit, uint32_t end4bit) const;

     uint32_t packRANS(uint32_t* in_packed, uint32_t start4bit, uint32_t end4bit) const;

    /**
     * Decodes number_of_out_elements packed half bytes to the byte array starting at out. Out will have half the size of the actual elements (since it is a vector of bytes instead of half bytes).
     */
    int unpackRANS(uint8_t *rans_begin, uint8_t *out, size_t number_of_output_elements) const;

    /**
     * Initializes iterative decoding for reading single elements from the decoding with itr_nextSymbol(). The state is carried in both of the parameters.
     */
    void itr_initDecoding(uint32_t* rans_state, const uint8_t** rans_ptr) const;
    void itr_initDecoding(uint32_t& rans_state, uint32_t& byte_index, const uint32_t* array) const;
    /**
     * Returns the next element from the decoding given the current internal state and updates the state in the two parameters.
     */
    uint32_t itr_nextSymbol(uint32_t* rans_state, const uint8_t** rans_ptr) const;
    uint32_t itr_nextSymbol(RansState& rans_state, uint32_t& byte_index,  const uint32_t* array) const;

    std::vector<uint32_t> getFrequencyArray() {
        assert(has_frequency_tables && "rANS instance has no frequency array");
        std::vector<uint32_t> out(16);
        for(int i=0; i < 16; i++)
            out[i] = stats.freqs[i];
        return out;
    }

    /**
     * @return an array string "uvec3(..), uvec3(..)" with 17 elements where each element < 16 is:
     * (dsys.start, dsyms.freq, stats.cum_freq) and element 16 is (0, 0, stats.cum_freq[16]).
     */
    std::string getGLSLSymbolArrayString() {
        // output our frequency tables:
        std::stringstream ss;
        for (int i = 0; i < RANS_ALPHABET_SIZE; i++) {
            ss << "uvec3(" << dsyms[i].start << "," << dsyms[i].freq << "," << stats.cum_freqs[i] << "),";
        }
        ss << "uvec3(0,0," << stats.cum_freqs[RANS_ALPHABET_SIZE] << ")";
        return ss.str();
    }


};

} // namespace vvv
