#ifndef COMPRESSED_SEGMENTATION_VOLUME_GLSL
#define COMPRESSED_SEGMENTATION_VOLUME_GLSL

// compile time macros:
//   CSVOL_ENCODING_ARRAY name of the SSBO from which the compressed bricks are read
//   CSVOL_DECODING_ARRAY name of the SSBO into which the decompressed bricks are written
//   USE_RANS                must be defined if and only if rANS encoding is used in the compression
//   USE_DETAIL_FREQ         must be defined if and only if the finest LoD uses another frequency table
//   RANS_SYMBOL_TABLE       array init. notation for 34 vec3 containing the freq. tables (see rans.glsl!)
//
#ifndef CSVOL_ENCODING_ARRAY
 #define CSVOL_ENCODING_ARRAY g_encoding
#endif
#ifndef CSVOL_DECODING_ARRAY
 #define CSVOL_DECODING_ARRAY g_decoding
#endif

// 4 bit operation symbols:
#include "cpp_glsl_include/csvol_constants.h"

#include "morton.glsl"
#ifdef USE_RANS
    #include "rans.glsl"
#endif

uint _indexOfBrickPos(const uvec3 p) {
    return morton3Dp2i(p);
}

// returns the label for the voxel position within the brick starting at the start_idx.
// decoded_inv_lod: the state of the brick in CSVOL_DECODING_ARRAY *must* be a full decoding up to this inv_lod
// brick_voxel: the coordinate of the lookup voxel on the *finest* lod, even if the lookup is for a coarser lod
uint readCSVolBrick(const uvec3 brick_voxel, const uint decoded_inv_lod, const uint decoded_brick_start_idx) {
    uint lod_width = g_brick_size / (1u << decoded_inv_lod);
    // determine position in brick
    return CSVOL_DECODING_ARRAY[decoded_brick_start_idx + (_indexOfBrickPos(brick_voxel) / (lod_width * lod_width * lod_width))];
}

const ivec3 neighbor[8][3] = {  {ivec3(-1, 0, 0), ivec3(0, -1, 0), ivec3(0, 0, -1)},
                                {ivec3( 1, 0, 0), ivec3(0, -1, 0), ivec3(0, 0, -1)},
                                {ivec3(-1, 0, 0), ivec3(0,  1, 0), ivec3(0, 0, -1)},
                                {ivec3( 1, 0, 0), ivec3(0,  1, 0), ivec3(0, 0, -1)},
                                {ivec3(-1, 0, 0), ivec3(0, -1, 0), ivec3(0, 0,  1)},
                                {ivec3( 1, 0, 0), ivec3(0, -1, 0), ivec3(0, 0,  1)},
                                {ivec3(-1, 0, 0), ivec3(0,  1, 0), ivec3(0, 0,  1)},
                                {ivec3( 1, 0, 0), ivec3(0,  1, 0), ivec3(0, 0,  1)}};

uint _unpack4BitFromEncoding(uint start, uint entry_id) {
    return bitfieldExtract(CSVOL_ENCODING_ARRAY[start + entry_id/8], 28 - int(entry_id % 8u) * 4, 4);
}

struct DiffTreeReadState {
    uint idxE;
    uint rans_state;
    bool in_detail_lod;
};


uint _readNextLodOperationFromEncoding(uint brick_start, inout DiffTreeReadState state) {
#ifdef USE_RANS
    return rans_itr_nextSymbol(state.rans_state, state.idxE, state.in_detail_lod);
#else
    return _unpack4BitFromEncoding(brick_start, state.idxE++);
#endif
}

uvec3 _enumBrickPos(uint i) {
    return morton3Di2p(i);
}


uint _valueOfNeighbor(const uvec3 brick_pos, const uint local_lod_i, const uint lod_width, const int neighbor_i, const uint decoded_brick_start_idx) {
    // find the position of the neighbor
    ivec3 neighbor_pos = ivec3(brick_pos) + neighbor[local_lod_i][neighbor_i] * int(lod_width);
    // find the index of the neighbor within the brick array
    uint neighbor_index = _indexOfBrickPos(uvec3(neighbor_pos));

    // in case we want to access a neighbor that is not already existing on this level, we have to
    // round down to the parent element of this element (lod_width*8)
    if(any(greaterThan(neighbor[local_lod_i][neighbor_i], ivec3(0))))
        neighbor_index -= neighbor_index % (lod_width * lod_width * lod_width * 8);

    // return value of neighbor or parent neighbor in brick
    return CSVOL_DECODING_ARRAY[decoded_brick_start_idx + neighbor_index];
}

// reset the brick by setting all of its entries to INVALID
void resetCSVolBrick(const uint decoded_brick_start_idx, const uint inv_lod) {
    uint voxel_count = 1u << (3u * inv_lod);
    for(uint i = decoded_brick_start_idx; i < decoded_brick_start_idx + voxel_count; i++) {
        CSVOL_DECODING_ARRAY[i] = INVALID;
    }
}

// fills the brick by setting all of its entries to value
void fillCSVolBrick(const uint decoded_brick_start_idx, const uint inv_lod, const uint value) {
    uint voxel_count = 1u << (3u * inv_lod);
    for(uint i = decoded_brick_start_idx; i < decoded_brick_start_idx + voxel_count; i++) {
        CSVOL_DECODING_ARRAY[i] = value;
    }
}

// Decompresses the encoding of the brick from the CSVOL_ENCODING_ARRAY array starting at encoding_start_index and ending
// at encoding_end_index up to the given inverse LOD level into the cached decompressed brick in CSVOL_DECODING_ARRAY.
// decoded_brick_start_index denotes the start position of this output brick.
// If start_at_inv_lod == 0, it is assumed that the output brick cache is set to INVALID at all entries.
// If start_at_inv_lod > 0, it is assumed that the output brick cache is fully decoded up to (start_at_inv_lod-1).
// start_at_inv_Lod mustn't already be the finest possible LoD.
void decompressCSVolBrick(const uint encoding_start_index, const uint encoding_end_index, const uvec3 valid_brick_size, const uint start_at_inv_lod, const uint inv_lod, const uint decoded_brick_start_idx) {
    // Refine up to the LOD that was requested, starting with decoding of start_at_inv_lod.
    // The starting position (taken from brick header) of the current LOD in the encoding array, measured in elements of uint.
    uint local_lod_i;
    // the palette starts at the end of the encoding block
    uint paletteE = encoding_end_index - 1u;
    uint beginE = encoding_start_index;

    DiffTreeReadState readState;    // read and changed in the _readNextLodOperationFromEncoding function
    readState.idxE = CSVOL_ENCODING_ARRAY[beginE + start_at_inv_lod];  // offset of current 4 bit entry to read
    readState.in_detail_lod = false;
#ifdef USE_RANS
    readState.idxE = (encoding_start_index + readState.idxE / 8u) * 4u;
    rans_itr_initDecoding(readState.rans_state, readState.idxE);
#endif

    uint output_size = (1u << inv_lod); // the resolution of voxel in each dimension of the output brick = for this LoD. On the finest LoD this is g_brick_size
    uint lod_width = (1u << inv_lod) / (1u << start_at_inv_lod);
    uint index_step = (lod_width * lod_width * lod_width);
    uint parent_value;

    // WE ASSUME that the brick was already completely set to INVALID
    //    // otherwise, set the whole brick to INVALID so we know later, what elements and LOD blocks were already processed
    //    for(uint i = 0; i < g_brick_size * g_brick_size * g_brick_size; i++)
    //        CSVOL_DECODING_ARRAY[i] = INVALID;

    for(uint lod = start_at_inv_lod; lod <= inv_lod; lod++) {
#ifdef USE_DETAIL_FREQ
#ifdef USE_RANS
        if(lod == g_lod_count - 1u) {
            readState.idxE = CSVOL_ENCODING_ARRAY[beginE + g_lod_count - 1u];
            readState.in_detail_lod = true;
            // a new rANS encoding starts for the finest LoD
            rans_itr_initDecoding(readState.rans_state, readState.idxE);
        }
#endif
#endif

        for (uint i = 0u; i < output_size * output_size * output_size; i += index_step) {
            // if an LOD block is completely outside the volume (i.e. it's first element is not within the volume)
            // we skip it as it won't have any entries in the encoding
            if (any(greaterThanEqual(_enumBrickPos(i).xyz * (g_brick_size/output_size), valid_brick_size)))
                continue;

            uint out_i = decoded_brick_start_idx + i;

            // every 8th element (we span 2*2*2=8 elements of the coarse LOD above), we fetch the new parent
            local_lod_i = (i % (index_step*8))/index_step;
            if (lod > 0u && i % (index_step*8) == 0) {
                // if this subtree is already filled, the last element of this block is set and we can skip it
                if (CSVOL_DECODING_ARRAY[out_i + (index_step*7)] != INVALID) {
                    i += (index_step*7);
                    continue;
                }

                parent_value = CSVOL_DECODING_ARRAY[out_i];
            }

            // get the next operation and apply it
            uint operation = _readNextLodOperationFromEncoding(beginE, readState);

            uint code_lsb = operation & 7u; // extract least significant 3 bits with 0111
            if (code_lsb == PARENT)
                CSVOL_DECODING_ARRAY[out_i] = parent_value;
            else if (code_lsb == NEIGHBORX)
                CSVOL_DECODING_ARRAY[out_i] = _valueOfNeighbor(_enumBrickPos(i), local_lod_i, lod_width, 0, decoded_brick_start_idx);
            else if (code_lsb == NEIGHBORY)
                CSVOL_DECODING_ARRAY[out_i] = _valueOfNeighbor(_enumBrickPos(i), local_lod_i, lod_width, 1, decoded_brick_start_idx);
            else if (code_lsb == NEIGHBORZ)
                CSVOL_DECODING_ARRAY[out_i] = _valueOfNeighbor(_enumBrickPos(i), local_lod_i, lod_width, 2, decoded_brick_start_idx);
            else if (code_lsb == PALETTE_ADV) {   // read palette entry and advance palette pointer to the next entry
                CSVOL_DECODING_ARRAY[out_i] = CSVOL_ENCODING_ARRAY[paletteE--];
            }
            else if (code_lsb == PALETTE_LAST) { // reuse the last palette entry
                CSVOL_DECODING_ARRAY[out_i] = CSVOL_ENCODING_ARRAY[paletteE+1];
            }
            else if (code_lsb == PALETTE_DELTA) {
                uint palette_delta = _readNextLodOperationFromEncoding(beginE, readState) + 2u;
                CSVOL_DECODING_ARRAY[out_i] = CSVOL_ENCODING_ARRAY[paletteE + palette_delta];
            }

            // stop traversal: fill all other parts of the brick with this value
            if ((operation & STOP_BIT) > 0u) {
                // fill the whole subtree with the parent value
                for (uint n = out_i; n < out_i + index_step; n++) {
                    CSVOL_DECODING_ARRAY[n] = CSVOL_DECODING_ARRAY[out_i];
                }
            }
        }
        // move to the next LOD block with half the block width and an eight of the index_step respectively
        index_step /= 8u;
        lod_width /= 2u;
    }
}

#endif /* COMPRESSED_SEGMENTATION_VOLUME_GLSL */
