#ifndef RANS_GLSL
#define RANS_GLSL

// adapted form the public domain ryg_rans rANS code by Fabian Giesen.
// see README.txt in include/rANS for details.

#define RANS_PROB_BITS 14
#define RANS_BYTE_L (1u << 23)  // lower bound of our normalization interval

// RANS_SYMBOL_TABLE must be defined at compile time to initilaize an array of uvec3 elements.
// for both the base encoding + for the detail back to back:
// decoding info for all 16 possible symbols: (dsyms.start, dsyms.freq, stats.cum_freq)
//                   + a 17th dummy entry with (0, 0, total stats.cum_freq)
// example:
// RANS_SYMBOL_TABLE=uvec3[34](uvec3(0,4659,0),uvec3(4659,1450,4659),uvec3(6109,771,6109), ...)
//
const uvec3 _RANS_STATS[34] = RANS_SYMBOL_TABLE;

uint _RansDecGet(uint rans_state) {
    return rans_state & ((1u << RANS_PROB_BITS) - 1u);
}

void _RansDecAdvanceSymbol(inout uint r, inout uint byte_index, uint sym_start, uint sym_freq, bool detail) {
    const uint mask = (1u << RANS_PROB_BITS) - 1u;

    // s, x = D(x)
    uint x = r;
    x = sym_freq * (x >> RANS_PROB_BITS) + (x & mask) - sym_start;

    // renormalize
    if (x < RANS_BYTE_L) {
        uint idx = byte_index;
        do {
            // read the next byte from our uint32 array
            uint shift = 8 * (idx % 4);

            // if the finest LoD is stored in a separate buffer, e.g. to stream it from CPU to GPU on demand,
            // this would be the place to access a different array:
            // uint byte = (detail ? (CSVOL_DETAIL_ARRAY[idx / 4] >> shift) : (CSVOL_ENCODING_ARRAY[idx / 4] >> shift)) & 0xFFu;
            // Additionally, rans_itr_initDecoding has to be similarly adapted.
            uint byte = (CSVOL_ENCODING_ARRAY[idx / 4] >> shift) & 0xFFu;       // alternatively, use bitfieldExtract(..)

            x = (x << 8) | byte;
            idx++;
        } while (x < RANS_BYTE_L);
        byte_index = idx;
    }
    r = x;
}

void rans_itr_initDecoding(inout uint rans_state, inout uint byte_index) {
    rans_state = CSVOL_ENCODING_ARRAY[byte_index/4];
    byte_index += 4;
}

uint rans_itr_nextSymbol(inout uint rans_state, inout uint byte_index, bool use_detail_freq) {
    uint cumulative = _RansDecGet(rans_state);
#ifdef USE_DETAIL_FREQ
    uint freq_table_offset = use_detail_freq ? 17u : 0u; // use another rANS frequency table for the detail level
#else
    uint freq_table_offset = 0u;
#endif
    for(s=freq_table_offset; s < freq_table_offset + 16u; s++) {
        if(_RANS_STATS[s+1].z > cumulative)
            break;
    }
    _RansDecAdvanceSymbol(rans_state, byte_index, _RANS_STATS[s].x, _RANS_STATS[s].y, use_detail_freq);
    return s - freq_table_offset;
}