#include "CompressedSegmentationVolume.hpp"

#include <algorithm>
#include <map>
#include <sstream>
#include <string>
#include <thread>
#include <iostream>

#include "bitpack.hpp"

namespace vvv {


uint32_t CompressedSegmentationVolume::valueOfNeighbor(const MultiGridNode* grid, const MultiGridNode* parent_grid,
                                                       const glm::uvec3 &brick_pos, const uint32_t local_lod_i,
                                                       const uint32_t lod_dim, const int neighbor_i) {
    assert(local_lod_i >= 0 && local_lod_i < 8);
    // Find the position of the neighbor.
    glm::ivec3 neighbor_pos = glm::ivec3(brick_pos) + neighbor[local_lod_i][neighbor_i];
    if (glm::any(glm::lessThan(neighbor_pos, glm::ivec3(0))) || glm::any(glm::greaterThanEqual(neighbor_pos, glm::ivec3(static_cast<int>(lod_dim)))))
        return INVALID;

        // In case we want to access a neighbor that is not already existing on this level (neighbor_i > our_i) or
        // any element of neighbor[local_lod_i][neighbor_i] is positive, we have to look up the parent element.
    else if (glm::any(glm::greaterThan(neighbor[local_lod_i][neighbor_i], glm::ivec3(0)))) {
        return parent_grid[pos2idx( glm::ivec3(brick_pos/2u) + neighbor[local_lod_i][neighbor_i], glm::uvec3(lod_dim/2))].label;
    }
        // Otherwise, lookup the neighbor.
    else {
        return grid[pos2idx(neighbor_pos, glm::uvec3(lod_dim))].label;
    }
}

uint32_t CompressedSegmentationVolume::valueOfNeighbor(const uint32_t* brick, const glm::uvec3 &brick_pos, const uint32_t local_lod_i, const uint32_t lod_width, const uint32_t brick_size,
                                                       const int neighbor_i) {
    assert(lod_width > 0);
    assert(local_lod_i >= 0 && local_lod_i < 8);
    // Find the position of the neighbor.
    glm::ivec3 neighbor_pos = glm::ivec3(brick_pos) + neighbor[local_lod_i][neighbor_i] * static_cast<int>(lod_width);
    if (glm::any(glm::lessThan(neighbor_pos, glm::ivec3(0))) || glm::any(glm::greaterThanEqual(neighbor_pos, glm::ivec3(static_cast<int>(brick_size)))))
        return INVALID;
    uint32_t neighbor_index = indexOfBrickPos(glm::uvec3(neighbor_pos));

    // In case we want to access a neighbor that is not already existing on this level (neighbor_i > our_i) or
    // any element of neighbor[local_lod_i][neighbor_i] is positive, we have to look up the parent element.
    if (glm::any(glm::greaterThan(neighbor[local_lod_i][neighbor_i], glm::ivec3(0))))
        neighbor_index -= neighbor_index % (lod_width * lod_width * lod_width * 8);

    // Since we don't check here if we're out of bounds of the volume, it CAN happen that a value is INVALID in the
    // encoding in the decoding, such a neighbor should never be accessed.
    // assert(brick[neighbor_index] != INVALID);

    return brick[neighbor_index];
}



// COMPRESSION ---------------------------------------------------------------------------------------------------------

void CompressedSegmentationVolume::setCompressionOptions(bool use_rANS, bool use_detail_freq, const uint32_t* operation_frequencies,
                                                         const uint32_t* detail_operation_frequencies) {
    m_use_rANS = use_rANS;
    m_use_detail_freq = use_detail_freq;

    if(m_use_detail_freq && !m_use_rANS)
        throw std::runtime_error("detail freq. can only be used in combination with rANS!");

    if(m_use_rANS) {
        assert(operation_frequencies != nullptr && "4 bit operation frequencies must be given if using rANS!");
        // normalize the symbol frequencies
        std::vector<uint32_t> _operation_frequencies = normalizeOperationFrequencies(operation_frequencies);
        m_rans.recomputeFrequencyTables(_operation_frequencies.data());
        if(m_use_detail_freq) {
            assert(detail_operation_frequencies != nullptr && "4 bit detail operation frequencies must be given if using rANS and detail freq.!");
            std::vector<uint32_t> _detail_operation_frequencies = normalizeOperationFrequencies(
                    detail_operation_frequencies);
            m_detail_rans.recomputeFrequencyTables(_detail_operation_frequencies.data());
        }
    }
}

void CompressedSegmentationVolume::setCompressionOptions64(bool use_rANS, bool use_detail_freq, const size_t *operation_frequencies,
                                                           const size_t *detail_operation_frequencies) {
    m_use_rANS = use_rANS;
    m_use_detail_freq = use_detail_freq;

    if(m_use_detail_freq && !m_use_rANS)
        throw std::runtime_error("detail freq. can only be used in combination with rANS!");

    if(m_use_rANS) {
        assert(operation_frequencies != nullptr && "4 bit operation frequencies must be given if using rANS!");
        // normalize the symbol frequencies
        std::vector<uint32_t> _operation_frequencies = normalizeOperationFrequencies(operation_frequencies);
        m_rans.recomputeFrequencyTables(_operation_frequencies.data());
        if(m_use_detail_freq) {
            assert(detail_operation_frequencies != nullptr && "4 bit detail operation frequencies must be given if using rANS and detail freq.!");
            std::vector<uint32_t> _detail_operation_frequencies = normalizeOperationFrequencies(
                    detail_operation_frequencies);
            m_detail_rans.recomputeFrequencyTables(_detail_operation_frequencies.data());
        }
    }
}


void CompressedSegmentationVolume::compressBrickForFrequencyTable(const std::vector<uint32_t>& volume, size_t* brick_freq,
                                                                  glm::uvec3 start, glm::uvec3 volume_dim, bool detail_freq) const {
    std::vector<uint32_t> palette;
    palette.reserve(32);
    glm::uvec3 volume_pos, brick_pos;

    const uint32_t lod_count = getLodCountPerBrick();

    // We keep track of the current brick status from coarsest to finest level to determine the right operations,
    // basically do an implicit decoding while we're encoding.
    uint32_t parent_value;
    uint32_t value;
    uint32_t local_lod_i;

    // construct the multigrid on this brick that we want to represent in this encoding
    std::vector<MultiGridNode> multigrid;
    constructMultiGrid(multigrid, volume, volume_dim, start, m_brick_size);

    // we start with the coarsest LOD, which is always a PALETTE_ADV of the max occuring value in the whole brick
    // we handle this here because it allows us to skip some special handling (for example checking if the palette is empty) in the following loop
    // in theory, we could start with a finer level here too and skip the first levels (= Carsten's original idea)
    uint32_t muligrid_lod_start = multigrid.size() - 1;
    if (multigrid[muligrid_lod_start].constant_subregion) {
        brick_freq[PALETTE_ADV | STOP_BIT]++;
    }
    else {
        brick_freq[PALETTE_ADV]++;
    }
    palette.push_back(multigrid[muligrid_lod_start].label);

    // now we iteratively refine from coarse (8 elements in the brick) to finest (brick_size^3 elements in the brick) levels
    uint32_t current_inv_lod = 1u;
    for (uint32_t lod_width = m_brick_size / 2u; lod_width > 0u; lod_width /= 2u) {
        // in the multigrid, LODs are ordered from finest to coarsest, so we have to go through them in reverse.
        uint32_t lod_dim = (m_brick_size/lod_width);
        uint32_t parent_multigrid_lod_start = muligrid_lod_start;
        muligrid_lod_start -= lod_dim * lod_dim * lod_dim;

        for (uint32_t i = 0; i < m_brick_size * m_brick_size * m_brick_size; i += lod_width * lod_width * lod_width) {
            brick_pos = enumBrickPos(i, m_brick_size);
            volume_pos = start + brick_pos;
            if (glm::any(glm::greaterThanEqual(volume_pos, volume_dim)))
                continue;

            local_lod_i = (i % (lod_width * lod_width * lod_width * 8)) / (lod_width * lod_width * lod_width);
            if (local_lod_i == 0) {
                if (multigrid[parent_multigrid_lod_start + pos2idx(brick_pos / lod_width / 2u, glm::uvec3(lod_dim / 2u))].constant_subregion) {
                    i += (lod_width * lod_width * lod_width * 7);
                    continue;
                }
                parent_value = multigrid[parent_multigrid_lod_start + pos2idx(brick_pos / lod_width / 2u, glm::uvec3(lod_dim / 2u))].label;
                assert(parent_value != INVALID && "parent element in brick was not set in previous LoD!");
            }

            value = multigrid[muligrid_lod_start + pos2idx(brick_pos / lod_width, glm::uvec3(lod_dim))].label;
            assert(value != INVALID && "Original volume mustn't contain the INVALID magic value!");

            uint32_t operation = 0u;
            if (lod_width >= 1 && multigrid[muligrid_lod_start + pos2idx(brick_pos / lod_width, glm::uvec3(lod_dim))].constant_subregion) {
                operation = STOP_BIT;
            }
            if (value == parent_value)
                operation |= PARENT;
            else if (valueOfNeighbor(multigrid.data() + muligrid_lod_start, multigrid.data() + parent_multigrid_lod_start, brick_pos / lod_width, local_lod_i, lod_dim, 0) == value)
                operation |= NEIGHBORX;
            else if (valueOfNeighbor(multigrid.data() + muligrid_lod_start, multigrid.data() + parent_multigrid_lod_start, brick_pos / lod_width, local_lod_i, lod_dim, 1) == value)
                operation |= NEIGHBORY;
            else if (valueOfNeighbor(multigrid.data() + muligrid_lod_start, multigrid.data() + parent_multigrid_lod_start, brick_pos / lod_width, local_lod_i, lod_dim, 2) == value)
                operation |= NEIGHBORZ;
            else if (palette.back() == value) // reuse the last palette value. but detail can't reference previous palettes!
                operation |= PALETTE_LAST;
            else {
                uint32_t palette_delta = static_cast<uint32_t>(std::find(palette.rbegin(), palette.rend(), value) - palette.rbegin());
                if(palette_delta < 17u && palette_delta < palette.size()) {
                    assert(palette.at(palette.size() - palette_delta - 1u) == value && "Palette value does not fit!");
                    assert(palette_delta > 0u &&
                           "the palette delta 0 should've been catched by the palette_last value!");
                    assert(palette_delta < 17u && "the palette delta can't be >= 17");
                    if (detail_freq && (current_inv_lod == lod_count - 1u))
                        brick_freq[16 + (operation | PALETTE_DELTA)]++;
                    else
                        brick_freq[operation | PALETTE_DELTA]++;
                    operation = palette_delta - 1u;
                }
                else {
                    palette.push_back(value);
                    operation |= PALETTE_ADV;
                }
            }
            assert(operation < 16u && "we only allow writing 4 bit operations!");
            if(detail_freq && (current_inv_lod == lod_count - 1u))
                brick_freq[16 + operation]++;
            else
                brick_freq[operation]++;

            assert(value != INVALID);
        }
        current_inv_lod++;
    }
}


void CompressedSegmentationVolume::compressForFrequencyTable(const std::vector<uint32_t>& volume, glm::uvec3 volume_dim, size_t freq_out[32], uint32_t subsampling_factor, bool detail_freq, bool verbose) {
    assert(m_brick_size > 0u && "brick size must be a power of 2 > 0");
    m_volume_dim = volume_dim;
    glm::uvec3 brickCount = getBrickCount();
    if(verbose) {
        std::cout << " running with " << NUM_THREADS << " threads on " << std::thread::hardware_concurrency() << " CPU cores" << std::endl;
        std::cout << " brick count: " << str(brickCount) << " = " << (brickCount.x * brickCount.y * brickCount.z)
                  << " with brick size " << m_brick_size << "^3 and " << str(m_volume_dim) << " voxels." << std::endl;
    }

    std::cout << " Progress 0.0%" << "\r" << std::flush;
    MiniTimer progressTimer;
    MiniTimer totalTimer;
    int bricks_since_last_update = 0;

    size_t brick_freq[NUM_THREADS][32]; // last 16 elements are detail frequencies, if detail separation is used
    for (int thread_id = 0; thread_id < NUM_THREADS; thread_id++) {
        for(int i = 0; i < 32; i++) {
            brick_freq[thread_id][i] = 0ul;
        }
    }

    glm::uvec3 brick;
    size_t brick_idx = 0;
    for (brick.z = 0u; brick.z < brickCount.z; brick.z+=subsampling_factor) {
        for (brick.y = 0u; brick.y < brickCount.y; brick.y+=subsampling_factor) {
            for (brick.x = 0u; brick.x < brickCount.x; brick.x+=(subsampling_factor * NUM_THREADS)) {

                #pragma omp parallel num_threads(NUM_THREADS) default(none) shared(brick, brickCount, volume, brick_freq, subsampling_factor, detail_freq)
                {
                    unsigned int thread_id = omp_get_thread_num();
                    if (brick.x + thread_id*subsampling_factor < brickCount.x) {

                        //volume, size_t* brick_freq, glm::uvec3 start, glm::uvec3 volume_dim, bool detail_freq
                        compressBrickForFrequencyTable(volume, brick_freq[thread_id],
                                   glm::uvec3(brick.x + thread_id * subsampling_factor, brick.y, brick.z) * m_brick_size,
                                   m_volume_dim, detail_freq);
                    }
                }

                // Progress output.
                bricks_since_last_update += NUM_THREADS;
                constexpr const double PROGRESS_UPDATE_INTERVAL = 2.;
                if (progressTimer.elapsed() >= PROGRESS_UPDATE_INTERVAL) {
                    float bricks_per_second = static_cast<float>(bricks_since_last_update) / progressTimer.elapsed();
                    std::stringstream stream;
                    stream << " Progress " << std::fixed << std::setprecision(1) << static_cast<float>(brick_idx) / static_cast<float>(brickCount.x * brickCount.y * brickCount.z / subsampling_factor / subsampling_factor / subsampling_factor) * 100.f << "%"
                           << " (" << std::setprecision(2) << (bricks_per_second * m_brick_size * m_brick_size * m_brick_size / 1000000.f)
                           << " million voxels/second)";
                    std::cout << stream.str() << "\r" << std::endl;
                    progressTimer.restart();
                    bricks_since_last_update = 0;
                }
            }
        }
    }


    // sum up the frequencies
    #pragma omp parallel for default(none) shared(freq_out, brick_freq)
    for(int i = 0; i < 32; i++) {
        freq_out[i] = 0ul;
        for (int thread_id = 0; thread_id < NUM_THREADS; thread_id++) {
            freq_out[i] += brick_freq[thread_id][i];
        }
    }

    float total_seconds = static_cast<float>(totalTimer.elapsed());
    std::cout << " Progress 100% in " << std::fixed << std::setprecision(3) << total_seconds
              << "s freq: " << array_string(freq_out, 16)  << " | " <<  array_string(freq_out + 16, 16) << std::endl;
}


std::vector<uint32_t> CompressedSegmentationVolume::normalizeOperationFrequencies(const size_t* freq) {
    std::vector<uint32_t> out(16);
    size_t operation_freq_sum = 0ul;
    for(int i =0; i < 16; i++) {
        operation_freq_sum += freq[i];
    }
    for (int i = 0; i < 16; i++) {
        bool greaterThanZero = freq[i] > 0u;
        out[i] = static_cast<uint32_t>(freq[i] / (operation_freq_sum  / (1ul << 30u) + 1u));
        if(greaterThanZero && out[i] == 0u)
            out[i] = 1u;        // existing symbols mustn't have a zero frequency
    }
    return out;
}

std::vector<uint32_t> CompressedSegmentationVolume::normalizeOperationFrequencies(const uint32_t* freq) {
    size_t f[16];
    for(int i = 0; i < 16; i++)
        f[i] = freq[i];
    return normalizeOperationFrequencies(f);
}

uint32_t CompressedSegmentationVolume::packRANS(std::vector<uint32_t>& entries, uint32_t start4bit, uint32_t end4bit, RANS& rANS) {
    return rANS.packRANS(entries, start4bit, end4bit);
}

uint32_t CompressedSegmentationVolume::compressBrick(const std::vector<uint32_t>& volume, std::vector<uint32_t>& out, glm::uvec3 start, glm::uvec3 volume_dim) {

    // BRICK MEMORY LAYOUT for L = log2(brick_size) LoDs
    // HEADER                 ENCODING:                              ┌ padded to 32bit
    // 4bit_encoding_start[0, 1, .. L-1], palette_start[0, 1 .. L], 4bit_encoding[0, 1, .. L], 32bit_palette[L, .. 1, 0]
    //       header_size*8 ᒧ                always zero ᒧ  ∟ .. one  ∟ palette size

    std::vector<uint32_t> palette;
    palette.reserve(32);
    glm::uvec3 volume_pos, brick_pos;

    const uint32_t lod_count = getLodCountPerBrick();
    const uint32_t header_size = 2u * lod_count + 1u;
    uint32_t out_i = header_size * 8u;  // write head position in out, counted as number of encoded 4 bit elements

    // We keep track of the current brick status from coarsest to finest level to determine the right operations,
    // basically do an implicit decoding while we're encoding.
    uint32_t parent_value;
    uint32_t value;
    uint32_t local_lod_i; // the local index of this element within the lod block of the coarser parent element, in 0 - 7, used for parent_value and neighbor-lookup index

    // construct the multigrid on this brick that we want to represent in this encoding
    std::vector<MultiGridNode> multigrid;
    constructMultiGrid(multigrid, volume, volume_dim, start, m_brick_size);

    // We start with the coarsest LOD, which is always a PALETTE_ADV of the max occuring value in the whole brick.
    // We handle this here because it allows us to skip some special handling (for example checking if the palette is empty) in the following loop.
    out[0] = out_i;                                     // inv. LoD 0 start position counted as 4bit elements
    out[header_size - 1] = 0u;                          // palette start position (from back)
    uint32_t muligrid_lod_start = multigrid.size() - 1; // coarsest LoD is the last entry of the multigrid
    if (multigrid[muligrid_lod_start].constant_subregion) {
        write4Bit(out, 0u, out_i++, PALETTE_ADV | STOP_BIT);
    }
    else {
        write4Bit(out, 0u, out_i++, PALETTE_ADV);
    }
    palette.push_back(multigrid[muligrid_lod_start].label);

    // now we iteratively refine from coarse (8 elements in the brick) to finest (brick_size^3 elements in the brick) levels
    uint32_t current_inv_lod = 1u;
    for (uint32_t lod_width = m_brick_size / 2u; lod_width > 0u; lod_width /= 2u) {
        // Write to header: where the new LODs start as a number of 4bit elements in encoding.
        out[current_inv_lod] = out_i;
        // Write to header: where the palette entries for the new LoD start.
        out[lod_count + current_inv_lod] = static_cast<uint32_t>(palette.size());

        // In the multigrid, LoDs are ordered from finest to coarsest, so we have to go through them in reverse.
        uint32_t lod_dim = (m_brick_size/lod_width);
        uint32_t parent_multigrid_lod_start = muligrid_lod_start;
        muligrid_lod_start -= lod_dim * lod_dim * lod_dim;

        for (uint32_t i = 0; i < m_brick_size * m_brick_size * m_brick_size; i += lod_width * lod_width * lod_width) {
            brick_pos = enumBrickPos(i, m_brick_size);
            // We don't store any operations for areas that would lie completely outside the volume.
            // The alternative would be to output a dummy value here (e.g. PARENT | STOP_BIT).
            volume_pos = start + brick_pos;
            if (glm::any(glm::greaterThanEqual(volume_pos, volume_dim)))
                continue;

            // Every 8th element, we fetch the new parent. 2*2*2=8 elements here have the same parent in the coarser LoD.
            local_lod_i = (i % (lod_width * lod_width * lod_width * 8)) / (lod_width * lod_width * lod_width);
            if (local_lod_i == 0) {
                // If our parent had a stop bit set (i.e. the multigrid parent indicated a constant subregion), all
                // current child elements are skipped. Note that this will also happen if this LoD block lies completely
                // outside the volume because a parent would've indicated a constant subregion earlier. Our parent
                // spanned 8 elements of this finer current level, so we advance 7 elements further.
                if (multigrid[parent_multigrid_lod_start + pos2idx(brick_pos / lod_width / 2u, glm::uvec3(lod_dim / 2u))].constant_subregion) {
                    i += (lod_width * lod_width * lod_width * 7);
                    continue;
                }

                // Otherwise, we remember the label of our parent for the parent operations that may follow.
                parent_value = multigrid[parent_multigrid_lod_start + pos2idx(brick_pos / lod_width / 2u, glm::uvec3(lod_dim / 2u))].label;
                assert(parent_value != INVALID && "parent element in brick was not set in previous LoD!");
            }

            value = multigrid[muligrid_lod_start + pos2idx(brick_pos / lod_width, glm::uvec3(lod_dim))].label;
            assert(value != INVALID && "Original volume mustn't contain the INVALID magic value!");

            // Find the best operation to encode the current multigrid node's label.
            uint32_t operation = 0u;
            // if the whole subtree from here has this parent_value, we can set a stop sign and fill the whole brick area of the subtree
            // note that LOD blocks outside the volume are by definition also homogeneous
            if (multigrid[muligrid_lod_start + pos2idx(brick_pos / lod_width, glm::uvec3(lod_dim))].constant_subregion) {
                operation = STOP_BIT;
            }
            if (value == parent_value)
                operation |= PARENT;
            else if (valueOfNeighbor(multigrid.data() + muligrid_lod_start, multigrid.data() + parent_multigrid_lod_start, brick_pos / lod_width, local_lod_i, lod_dim, 0) == value)
                operation |= NEIGHBORX;
            else if (valueOfNeighbor(multigrid.data() + muligrid_lod_start, multigrid.data() + parent_multigrid_lod_start, brick_pos / lod_width, local_lod_i, lod_dim, 1) == value)
                operation |= NEIGHBORY;
            else if (valueOfNeighbor(multigrid.data() + muligrid_lod_start, multigrid.data() + parent_multigrid_lod_start, brick_pos / lod_width, local_lod_i, lod_dim, 2) == value)
                operation |= NEIGHBORZ;
            else if (palette.back() == value)
                operation |= PALETTE_LAST;
            else {
                // reuse the n-delta palette value when 0 < delta < 17
                uint32_t palette_delta = static_cast<uint32_t>(std::find(palette.rbegin(), palette.rend(), value) - palette.rbegin());
                if(palette_delta < 17u && palette_delta < palette.size()) {
                    assert(palette.at(palette.size() - palette_delta - 1u) == value && "Palette value does not fit!");
                    assert(palette_delta > 0u && "the palette delta 0 should've been catched by the palette_last value!");
                    assert(palette_delta < 17u && "the palette delta can't be >= 17");

                    // we write two "operations". First, the actual operation for the palette backwards reference:
                    write4Bit(out, 0u, out_i++, operation | PALETTE_DELTA);
                    // Second, the delta of the reference:
                    // The "0" case is already handled by PALETTE_LAST, so we only consider case 1 - 16 in our 4 bits.
                    operation = palette_delta - 1u;
                }
                    // if nothing helps, we add a completely new palette entry
                else {
                    palette.push_back(value);
                    operation |= PALETTE_ADV;
                }
            }
            assert(operation < 16u && "We only allow writing 4 bit operations!");
            write4Bit(out, 0u, out_i++, operation);
        }


        // we handle the rANS packing here, if we're using separate frequency tables for the finest LoD and the coarser
        // LoDs. Otherwise, we will do the rANS packing with the default encoder after all loops finished execution.
        if(m_use_detail_freq) {
            // Pack all previous levels via rANS encoding if we're at the second last LoD.
            // This is the last LoD of non-detail encoding for which another rANS encoder will be used.
            if (current_inv_lod == lod_count - 2u) {
                out_i = packRANS(out, out[0], out_i, m_rans);
                assert(out_i % 8u == 0u && "Next element after rANS output should start at a full uint32_t element!");
            }
                // Pack the detail (=finest LoD) via rANS encoding if we're at the finest LoD.
                // We use a separate rANS encoder here because the detail level doesn't use stop bits => different operation frequencies
            else if (current_inv_lod == lod_count - 1u) {
                out_i = packRANS(out, out[current_inv_lod], out_i, m_detail_rans);
            }
        }

        current_inv_lod++;
    }
    // Pack everything via rANS encoding if we didn't pack it before.
    // i.e. when using the same rANS encoder for all LoDs.
    if(!m_use_detail_freq && m_use_rANS)
        out_i = packRANS(out, out[0], out_i, m_rans);

    // last entry of our header stores the palette size
    out[header_size - 1u] = palette.size();

    // After the encoding follows the palette containing the labels.
    // Round up to start the palette at a full uint32_t index, but AFTER the last encoding element.
    while(out_i % 8u != 0u)
        write4Bit(out, 0u, out_i++, 0u);
    // Now we count everything in 32 bit elements:
    out_i /= 8u;
    // The palette is added in reverse order at the end to be read from the full compressed block back to front.
    for(int i = static_cast<int>(palette.size()) - 1; i >= 0; i--) {
        out.at(out_i++) = palette.at(i);
    }

    assert(out_i < out.size() && "out doesn't provide enough memory for encoded brick, writing outside of allocated region");
    return out_i; // Return the number of uint32_t elements that we wrote for this block.
}

void CompressedSegmentationVolume::compress(const std::vector<uint32_t> &volume, const glm::uvec3 volume_dim, bool verbose) {
    assert(m_brick_size > 0u && "brick size must be a power of 2 > 0");
    m_volume_dim = volume_dim;
    glm::uvec3 brickCount = getBrickCount();
    if(verbose) {
        std::cout << " running with " << NUM_THREADS << " threads on " << std::thread::hardware_concurrency() << " CPU cores" << std::endl;
        std::cout << " brick count: " << str(brickCount) << " = " << (brickCount.x * brickCount.y * brickCount.z) << " with brick size " << m_brick_size << "^3" << std::endl;
    }

    m_encoding.clear();
    m_encoding.reserve(m_volume_dim.x * m_volume_dim.y * m_volume_dim.z / 12u / 7); // assume that we have a compression rate below 1/12
    m_brick_starts.resize(brickCount.x * brickCount.y * brickCount.z + 1);

    // Tracking the compression progress.
    if(verbose)
        std::cout << " Progress 0.0%" << "\r" << std::flush;
    MiniTimer progressTimer;
    MiniTimer totalTimer;
    int bricks_since_last_update = 0;

    // compute the next NUM_THREADS brick encodings in parallel.
    // Multiple buffers are required to store the temporary results of compressing the single bricks.
    std::vector<uint32_t> encodedBrick[NUM_THREADS];
    uint32_t encoded_element_count[NUM_THREADS];
    // Assume a worst case compression rate of 100%.
    const uint32_t max_encoded_brick_count_32bit = m_brick_size * m_brick_size * m_brick_size;
    for (int thread_id = 0; thread_id < NUM_THREADS; thread_id++) {
        encodedBrick[thread_id].resize(max_encoded_brick_count_32bit / 4u);
        encoded_element_count[thread_id] = 0u;
    }

    glm::uvec3 brick;
    size_t brick_idx = 0;
    for (brick.z = 0u; brick.z < brickCount.z; brick.z++) {
        for (brick.y = 0u; brick.y < brickCount.y; brick.y++) {
            for (brick.x = 0u; brick.x < brickCount.x; brick.x += NUM_THREADS) {

                // Compress the current line of bricks in parallel.
                #pragma omp parallel num_threads(NUM_THREADS) default(none) shared(brick, brickCount, volume, encodedBrick, encoded_element_count)
                {
                    unsigned int thread_id = omp_get_thread_num();
                    encoded_element_count[thread_id] = 0u;
                    if (brick.x + thread_id < brickCount.x) {
                        encoded_element_count[thread_id] = compressBrick(volume, encodedBrick[thread_id],
                                                                         glm::uvec3(brick.x + thread_id, brick.y, brick.z)
                                                                         * m_brick_size, m_volume_dim);
                    }
                }

                // Append the results to the complete buffers.
                for (int thread_id = 0; thread_id < NUM_THREADS; thread_id++) {
                    if (encoded_element_count[thread_id] == 0u)
                        continue;
                    // Store the start index of the brick in the encoding array.
                    assert(brick_idx < m_brick_starts.size() && "out of bounds");
                    m_brick_starts[brick_idx++] = static_cast<uint32_t>(m_encoding.size());
                    // Store the encoded brick at the right position in the encoding array (ordered by brick index).
                    size_t old_size = m_encoding.size();
                    m_encoding.resize(old_size + encoded_element_count[thread_id]);
                    memcpy(m_encoding.data() + old_size, encodedBrick[thread_id].data(), encoded_element_count[thread_id] * sizeof(uint32_t));
                }


                // Output a progress update after some fixed time interval.
                if(verbose) {
                    bricks_since_last_update += NUM_THREADS;
                    constexpr const double PROGRESS_UPDATE_INTERVAL = 2.;
                    if (progressTimer.elapsed() >= PROGRESS_UPDATE_INTERVAL) {
                        float bricks_per_second = static_cast<float>(bricks_since_last_update) / progressTimer.elapsed();
                        float remaining_seconds = static_cast<float>(brickCount.x * brickCount.y * brickCount.z - brick_idx) / bricks_per_second;
                        std::stringstream stream;
                        stream << " Progress " << std::fixed << std::setprecision(1) << static_cast<float>(brick_idx) / static_cast<float>(brickCount.x * brickCount.y * brickCount.z) * 100.f << "%"
                               << " (" << std::setprecision(2) << (bricks_per_second * m_brick_size * m_brick_size * m_brick_size / 1000000.f)
                               << " million voxels/second), remaining: " << static_cast<int>(remaining_seconds / 60.f) << "m" << (static_cast<int>(remaining_seconds) % 60) << "s        ";
                        std::cout << stream.str() << "\r" << std::flush;
                        progressTimer.restart();
                        bricks_since_last_update = 0;
                    }
                }
            }
        }
    }

    // One last dummy entry to be able to query an "end" index for the last brick.
    m_brick_starts[brick_idx] = static_cast<uint32_t>(m_encoding.size());

    // Output a final progress update.
    float total_seconds = static_cast<float>(totalTimer.elapsed());
    std::cout << " ______________________________________________________________________________________" << std::endl;
    std::cout << " Progress 100% in " << std::fixed << std::setprecision(3) << total_seconds << "s ("
              << (static_cast<float>(volume.size()) / total_seconds / 1000000.f) << " million voxels/second) "
              << decodingInfoString() << std::endl;
}

// DECOMPRESSION ---------------------------------------------------------------------------------------------------

uint32_t CompressedSegmentationVolume::readNextLodOperation(size_t start32bit, CompressedSegmentationVolume::ReadState &state) const {
    if (m_use_rANS) {
        const RANS* rANS_decoder = (m_use_detail_freq && state.in_detail_lod) ? &m_detail_rans : &m_rans;
        return rANS_decoder->itr_nextSymbol(state.rans_state, state.idxE, m_encoding.data());
    }
    else
        return read4Bit(m_encoding, start32bit, state.idxE++);
}

void CompressedSegmentationVolume::decompressBrick(uint32_t brick_idx, uint32_t* output_brick, glm::uvec3 valid_brick_size, int inv_lod) const {
    // the palette starts at the end of the encoding block
    uint32_t beginE = m_brick_starts[brick_idx];
    uint32_t paletteE = m_brick_starts[brick_idx + 1u] - 1u;
    uint32_t lod_count = getLodCountPerBrick();

    // refine up to the inverse LoD that was requested
    uint32_t local_lod_i;   // local index within the block of elements that have the same coarser parent (in 0 - 7)

    ReadState readState = {.idxE=m_encoding[beginE], .in_detail_lod=false};
    if(m_use_rANS) {
        readState.idxE = (beginE + readState.idxE / 8) * 4;
        m_rans.itr_initDecoding(readState.rans_state, readState.idxE, m_encoding.data());
    }

    uint32_t index_step = m_brick_size * m_brick_size * m_brick_size;
    uint32_t lod_width = m_brick_size;
    uint32_t parent_value;

    // First, set the whole brick to INVALID to know later which elements and LoD blocks were already processed.
    for (uint32_t i = 0; i < m_brick_size * m_brick_size * m_brick_size; i++)
        output_brick[i] = INVALID;

    for (int lod = 0; lod <= inv_lod; lod++) {

        // in case we're using a different rANS frequency table for the finest LoD:
        // check if we ran into the detail layer (finest LoD) and change the readState accordingly
        if(m_use_detail_freq && lod == lod_count-1) {
            // Read the lod start from the brick header to start reading at the right encoding buffer index.
            // We have to start at a fully padded uint32, because we switch the rANS decoder.
            readState.idxE = (beginE + m_encoding[beginE + lod] / 8) * 4;
            readState.in_detail_lod = true;
            m_detail_rans.itr_initDecoding(readState.rans_state, readState.idxE, m_encoding.data());
        }

        for (uint32_t i = 0; i < m_brick_size * m_brick_size * m_brick_size; i += index_step) {
            // if an LOD block is completely outside the volume (i.e. it's first element is not within the volume) we skip it as it won't have any entries in the encoding
            if (glm::any(glm::greaterThanEqual(enumBrickPos(i, m_brick_size), valid_brick_size)))
                continue;

            // every 8th element (we span 2*2*2=8 elements of the coarse LOD above), we fetch the new parent
            local_lod_i = (i % (index_step * 8)) / index_step;
            if (lod > 0 && i % (index_step * 8) == 0) {
                // if the last element of this block is set, this subtree is already filled (because of a PARENT_STOP
                // operation in a coarser LoD before) and we can skip it.
                if (output_brick[i + (index_step * 7)] != INVALID) {
                    i += (index_step * 7);
                    continue;
                }

                parent_value = output_brick[i];
                assert(parent_value != INVALID && "Parent element in brick was not set in previous LoD!");
            }

            // get the next operation and apply it
            uint32_t operation = readNextLodOperation(beginE, readState);
            uint32_t operation_lsb = operation & 7u; // extract least significant 3 bits
            if (operation_lsb == PARENT)
                output_brick[i] = parent_value;
            else if (operation_lsb == NEIGHBORX)
                output_brick[i] = valueOfNeighbor(output_brick, enumBrickPos(i, m_brick_size), local_lod_i, lod_width, m_brick_size, 0);
            else if (operation_lsb == NEIGHBORY)
                output_brick[i] = valueOfNeighbor(output_brick, enumBrickPos(i, m_brick_size), local_lod_i, lod_width, m_brick_size, 1);
            else if (operation_lsb == NEIGHBORZ)
                output_brick[i] = valueOfNeighbor(output_brick, enumBrickPos(i, m_brick_size), local_lod_i, lod_width, m_brick_size, 2);
            else if (operation_lsb == PALETTE_ADV) { // read palette entry and advance palette pointer to the next entry
                output_brick[i] = m_encoding[paletteE--];
            }
            else if (operation_lsb == PALETTE_LAST) {
                output_brick[i] = m_encoding[paletteE + 1];
            }
            else if (operation_lsb == PALETTE_DELTA) {
                uint32_t palette_delta = readNextLodOperation(beginE, readState) + 2u;
                output_brick[i] = m_encoding[paletteE + palette_delta];
            }
            else
                assert(false && "unrecognized compression operation");

            // If the stop bit was set, stop traversal: fill all other parts of the brick with this value.
            if ((operation & STOP_BIT) != 0u) {
                for (uint32_t n = i; n < i + index_step; n++) {
                    output_brick[n] = output_brick[i];
                }
            }

            assert(output_brick[i] != INVALID && "Set output element brick to forbidden magic value INVALID!");
        }

        // move to the next LOD block with half the block width and an eight of the index_step respectively
        index_step /= 8;
        lod_width /= 2;
    }
}

void CompressedSegmentationVolume::decompressBrickTo(uint32_t* out, glm::uvec3 brick_pos, int target_lod) const {
    int inv_lod = static_cast<int>(getLodCountPerBrick()) - 1 - target_lod;
    const glm::uvec3 brickCount = getBrickCount();
    size_t brick_idx = brick_to_1D(brick_pos, brickCount);
    decompressBrick(brick_idx, out,
                    glm::clamp(m_volume_dim - brick_pos * m_brick_size, glm::uvec3(0u), glm::uvec3(m_brick_size)),
                    inv_lod);
}

void CompressedSegmentationVolume::decompressLoD(int lod, std::vector<uint32_t>& out) const {

    const glm::uvec3 brickCount = getBrickCount();
    int inv_lod = getLodCountPerBrick() - 1 - lod;
    assert(inv_lod >= 0);


    // For CPU decoding of a full volume up to a certain LoD, we assume that the voxels are stored in the same linearized
    // XYZ order as the original volume. To that end, we have to remap the ordering of the voxels using temporary buffers.
    // This behavior can be disabled by commenting out the following define, leading to a decompression where each brick
    // is stored in morton order (Z-order) in the output volume buffer back to back :
    //#define NO_BRICK_DECODE_INDEX_REMAP

    // this would run in parallel on the GPU later!
    glm::uvec3 brick_pos;
#ifndef NO_BRICK_DECODE_INDEX_REMAP
    std::vector<uint32_t> brick_cache[NUM_THREADS]; // Contains voxels in morton order
    for (auto &bc : brick_cache)
        bc.resize(m_brick_size * m_brick_size * m_brick_size);
#else
    void* brick_cache = nullptr; // just for OpenMP
#endif
#pragma omp parallel for num_threads(NUM_THREADS) default(none) private(brick_pos) shared(brickCount, brick_cache, out, inv_lod)
    for (uint32_t z = 0; z < brickCount.z; z++) {
        unsigned int thread_id = omp_get_thread_num();
        brick_pos.z = z;
        for (brick_pos.y = 0; brick_pos.y < brickCount.y; brick_pos.y++) {
            for (brick_pos.x = 0; brick_pos.x < brickCount.x; brick_pos.x++) {
                size_t brick_idx = brick_to_1D(brick_pos, brickCount);
#ifndef NO_BRICK_DECODE_INDEX_REMAP
                // Decode brick to a buffer where the voxels are linearized in morton order.
                decompressBrick(brick_idx, brick_cache[thread_id].data(),
                                glm::clamp(m_volume_dim - brick_pos * m_brick_size, glm::uvec3(0u),
                                           glm::uvec3(m_brick_size)),
                                inv_lod);
                // Fill the output array with decoded brick entries in linearized XYZ order.
                for (uint32_t i = 0; i < m_brick_size * m_brick_size * m_brick_size; i++) {
                    glm::uvec3 out_pos = brick_pos * m_brick_size + enumBrickPos(i, m_brick_size);
                    if (glm::all(glm::lessThan(out_pos, m_volume_dim))) {
                        out[pos2idx(out_pos, m_volume_dim)] = brick_cache[thread_id][i];
                    }
                }
#else
                decodeBrick(brick_idx, &(out[pos2idx(brick_pos * m_brick_size, m_volume_dim)]),
                            glm::clamp(m_volume_dim - brick_pos * m_brick_size, glm::uvec3(0u), glm::uvec3(m_brick_size)),
                            inv_lod);
#endif
            }
        }
    }
}


// TESTS -----------------------------------------------------------------------------------------------------------

bool CompressedSegmentationVolume::testLoD(const std::vector<uint32_t> &volume, glm::uvec3 volume_dim) const {
    assert(volume.size() == volume_dim.x * volume_dim.y * volume_dim.z && "Volume size does not match dimension!");

    std::cout << " Running LoD compression test" << std::endl;


    MiniTimer timer;
    static constexpr int max_error_lines = 8;
    size_t error_count = 0;
    bool allgood = true;
    // Reserve space to store a full volume:
    std::vector<uint32_t> out;
    out.resize(static_cast<size_t>(m_volume_dim.x) * static_cast<size_t>(m_volume_dim.y) * static_cast<size_t>(m_volume_dim.z));

    // Check the LODs over increasing LOD brick size. We skip LoD 0, because that's already tested in the 'test' method
    // of VolumeCompressionBase and is technically the volume without any LOD.
    int lod = 1;
    uint32_t multigrid_lod_start = m_brick_size * m_brick_size * m_brick_size;
    for (uint32_t width = 2; width <= m_brick_size; width *= 2) {
        timer.restart();
        std::cout << " Decompress LoD " << lod << " with block width " << width << "\r" << std::flush;
        decompressLoD(lod, out);
        std::cout << " Decompress LoD " << lod << " with block width " << width << " in " << timer.elapsed() << "s done. Test: ";
        if (volume.size() != out.size()) {
            std::cout << " ERROR: Compressed in and out sizes don't match" << std::endl;
            std::cout << " skipping other LODs..." << std::endl;
            std::cout << "-------------------------------------------------------------" << std::endl;
            return false;
        }

        // Iterate over all bricks but only check the currently tested LoD.
        error_count = 0;
        const glm::uvec3 brickCount = getBrickCount();
        glm::uvec3 brick;
        size_t brick_idx = 0;
        #pragma omp parallel for default(none) private(brick) shared(width, volume_dim, volume, out, error_count, max_error_lines, brickCount, multigrid_lod_start, std::cout)
        for (uint32_t z = 0u; z < brickCount.z; z++) {
            brick.z = z;
            for (brick.y = 0u; brick.y < brickCount.y; brick.y++) {
                for (brick.x = 0u; brick.x < brickCount.x; brick.x += NUM_THREADS) {

                    // Construct target multigrid for this brick (quite inefficient since we only test one level here).
                    std::vector<MultiGridNode> multigrid;
                    constructMultiGrid(multigrid, volume, volume_dim, brick * m_brick_size, m_brick_size);

                    // check all elements of this LoD
                    glm::uvec3 pos_in_brick;
                    for (pos_in_brick.z = 0; pos_in_brick.z < m_brick_size/width; pos_in_brick.z++) {
                        for (pos_in_brick.y = 0; pos_in_brick.y < m_brick_size/width; pos_in_brick.y += width) {
                            for (pos_in_brick.x = 0; pos_in_brick.x < m_brick_size/width; pos_in_brick.x += width) {
                                if(glm::any(glm::greaterThanEqual(brick * m_brick_size + pos_in_brick * width, volume_dim)))
                                    continue;

                                uint32_t i = pos2idx(brick * m_brick_size + pos_in_brick * width, volume_dim);
                                uint32_t expected_value = multigrid[multigrid_lod_start + pos2idx(pos_in_brick, glm::uvec3(m_brick_size/width))].label;

                                if (expected_value != out[i]) {
                                    #pragma omp critical
                                    {
                                        error_count++;
                                        if (error_count <= max_error_lines)
                                            std::cout << "error at " << str(idx2pos(i, volume_dim)) << " in " <<
                                                          volume[i] << " != out " << out[i] <<
                                                          " multigrid LoD start index " << multigrid_lod_start << std::endl;
                                        else if (error_count == max_error_lines + 1)
                                            std::cout << "[...] skipping additional errors" << std::endl;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }


        size_t lod_total_number_of_elements = ((volume_dim.x + width - 1) / width) * ((volume_dim.y + width - 1) / width) * ((volume_dim.z + width - 1) / width);
        std::cout << "finished with " << error_count << " / " << lod_total_number_of_elements << " errors ("
                     << (100.f * static_cast<float>(error_count) / static_cast<float>(lod_total_number_of_elements)) << "%)" << std::endl;

        allgood &= (error_count == 0);
        lod++;
        multigrid_lod_start += (m_brick_size / width) * (m_brick_size / width) * (m_brick_size / width);
    }

    if(allgood)
        std::cout << "no errors!" << std::endl;
    else
        std::cout << "encountered errors!" << std::endl;

    std::cout << "-------------------------------------------------------------" << std::endl;
    return error_count == 0;
}


// FILE IMPORT/EXPORT --------------------------------------------------------------------------------------------------

void CompressedSegmentationVolume::exportToFile(const std::string &path, bool verbose) {
    if (m_brick_starts.empty() || m_encoding.empty()) {
        std::cout << "Encoding was not yet computed. Call compress(..) first. Skipping." << std::endl;
        return;
    }
    if (std::filesystem::exists(path)) {
        std::cout << "File " << path << " already exist. Skipping." << std::endl;
        return;
    }
    std::ofstream file(path, std::ios_base::out | std::ios::binary);
    if (!file.is_open()) {
        std::cout << "Unable to open export file " << path << ". Skipping." << std::endl;
        return;
    }

    // write header: 8 chars CMPSGVOL + 4 chars version number
    const char *magic_header = "CMPSGVOL";
    const char *version = "0010"; // paper release version
    file.write(magic_header, 8);
    file.write(version, 4);
    // write general info
    file.write(reinterpret_cast<char *>(&m_brick_size), sizeof(uint32_t));
    file.write(reinterpret_cast<char *>(&m_volume_dim), sizeof(glm::uvec3));
    file.write(reinterpret_cast<char *>(&m_use_rANS), sizeof(bool)); // since 0002
    if(m_use_rANS) {
        auto freq_table = m_rans.getFrequencyArray();
        for(int i = 0; i < 16; i++)
            file.write(reinterpret_cast<char *>(&freq_table[i]), sizeof(uint32_t));
        file.write(reinterpret_cast<char *>(&m_use_detail_freq), sizeof(bool));
        if(m_use_detail_freq) {
            freq_table = m_detail_rans.getFrequencyArray();
            for(int i = 0; i < 16; i++)
                file.write(reinterpret_cast<char *>(&freq_table[i]), sizeof(uint32_t));
        }
    }
    // write brick starts buffer and encoding
    size_t size = m_brick_starts.size();
    file.write(reinterpret_cast<char *>(&size), sizeof(size_t));
    file.write(reinterpret_cast<char *>(&m_brick_starts[0]), static_cast<long>(size * sizeof(m_brick_starts[0])));
    size = m_encoding.size();
    file.write(reinterpret_cast<char *>(&size), sizeof(size_t));
    file.write(reinterpret_cast<char *>(&m_encoding[0]), static_cast<long>(size * sizeof(m_encoding[0])));
    file.close();
    if(verbose)
        std::cout << "Exported compressed segmentation volume to " << path << std::endl;
}

bool CompressedSegmentationVolume::importFromFile(const std::string &path, bool verbose) {
    std::ifstream fin(path, std::ios::in | std::ios::binary);
    if (!fin.is_open()) {
        if(verbose)
            std::cout << "Unable to open import file " << path << ". Skipping." << std::endl;
        return false;
    }
    // Check header and version.
    char magic_header[9];
    char _version[5];
    fin.read(reinterpret_cast<char *>(magic_header), 8);
    magic_header[8] = '\0';
    if (std::string(magic_header) != "CMPSGVOL") {
        std::cout << "File " << path << " is not a compressed segmentation volume. Missing header CMPSGVOL." << std::endl;
        return false;
    }
    fin.read(reinterpret_cast<char *>(_version), 4);
    _version[4] = '\0';
    std::string version(_version);
    if (version != "0010") {
        std::cout << "Import does not support version " << version << " of file " << path << ". Skipping." << std::endl;
        return false;
    }

    // Read the general data set info.
    fin.read(reinterpret_cast<char *>(&m_brick_size), sizeof(uint32_t));
    fin.read(reinterpret_cast<char *>(&m_volume_dim), sizeof(glm::uvec3));
    fin.read(reinterpret_cast<char *>(&m_use_rANS), sizeof(bool));
    if (m_use_rANS) {
        uint32_t freq_table[16];
        for (int i = 0; i < 16; i++)
            fin.read(reinterpret_cast<char *>(&freq_table[i]), sizeof(uint32_t));
        m_rans.recomputeFrequencyTables(freq_table);
        fin.read(reinterpret_cast<char *>(&m_use_detail_freq), sizeof(bool));
        if (m_use_detail_freq) {
            for (int i = 0; i < 16; i++)
                fin.read(reinterpret_cast<char *>(&freq_table[i]), sizeof(uint32_t));
            m_detail_rans.recomputeFrequencyTables(freq_table);
        }
    }
    // Read the data directly to our members.
    size_t size;
    fin.read(reinterpret_cast<char *>(&size), sizeof(size_t));
    m_brick_starts.resize(size);
    fin.read(reinterpret_cast<char *>(&m_brick_starts[0]), static_cast<long>(size * sizeof(uint32_t)));
    fin.read(reinterpret_cast<char *>(&size), sizeof(size_t));
    m_encoding.resize(size);
    fin.read(reinterpret_cast<char *>(&m_encoding[0]), static_cast<long>(size * sizeof(uint32_t)));
    fin.close();
    glm::uvec3 brickCount = getBrickCount();
    if(verbose)
        std::cout << "Imported CompressedSegmentationVolume from " << path << " with brick count: "
                  << str(brickCount) << " = " << (brickCount.x * brickCount.y * brickCount.z) << " with brick size "
                  << m_brick_size << "^3" << std::endl;
    return true;
}



}