#include <memory>
#include "CompressedSegmentationVolume.hpp"
#include "Volume.hpp"


int main(int argc, char *argv[]) {

    // CONFIGURATION ------------------
    const int brick_dim = 32;                           // size of one brick
    const bool use_rANS = true;                         // use secondary compression with rANS
    const bool use_detail_freq = true;                  // use a separate freq. table for brick leaf nodes / finest LoD
    const unsigned int frequency_pass_subsampling = 8u; // only use every n³th block for computing frequencies
    //
    const bool verbose = true;
    const int example_volume_dim = 128;                 // be careful with large volume sizes and label densities
    const int example_volume_label_density = 100;       // labels per million voxels
    // ---------------------------------


    // SEGMENTATION VOLUME IMPORT  -----
    // Try to load a precomputed uint32_t segmentation volume or create a new one.
    Volume volume;
    if(!volume.importFromFile("./example_segmentation_voronoi.csgv")) {
        // Careful with large volume sizes or label numbers!
        volume.createSyntheticSegmentationVolume(example_volume_dim, example_volume_dim, example_volume_dim,
                                                 example_volume_label_density);
        volume.exportToFile("./example_segmentation_voronoi.csgv");
        std::cout << std::endl;
    }
    glm::ivec3 volume_dim = glm::ivec3(volume.dim_x, volume.dim_y, volume.dim_z);;
    // ---------------------------------


    // COMPRESSION ---------------------
    // create an empty compressed segmentation volume object
    std::shared_ptr<vvv::CompressedSegmentationVolume> csvol = std::make_shared<vvv::CompressedSegmentationVolume>(brick_dim);
    // compute frequencies of operations in a compressed volume for later use in the rANS encoding
    size_t operation_frequencies[32];
    if(use_rANS) {
        std::cout << std::endl << "Operation frequency pass:" << std::endl;
        csvol->compressForFrequencyTable(volume.data(), volume_dim, operation_frequencies, frequency_pass_subsampling,
                                         use_detail_freq, false);
        // We can't risk missing any existing operations in our table due to subsampling,
        // so we set any frequency to at least one.
        if(frequency_pass_subsampling > 1u) {
            for(int i = 0; i < 16; i++) {
                operation_frequencies[i] = std::max(operation_frequencies[i], 1ul);
                if(use_detail_freq)
                    operation_frequencies[i + 16] = std::max(operation_frequencies[i + 16], 1ul);
            }
        }
    }
    // perform the compression
    std::cout << "Compression pass:                             " << std::endl;
    csvol->setCompressionOptions64(use_rANS, use_detail_freq, operation_frequencies, operation_frequencies + 16);
    csvol->compress(volume.data(), volume_dim, verbose);
    // ---------------------------------


    // DECOMPRESSION TESTS -------------
    std::cout << std::endl;
    csvol->test(volume.data(), volume_dim, false);
    // ---------------------------------


    return 0;
}